"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var UnifyClasses_1 = require("./UnifyClasses");
function unionMembersRecursively() {
    var unions = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        unions[_i] = arguments[_i];
    }
    var processedUnions = immutable_1.Set();
    var members = immutable_1.OrderedSet();
    function addMembers(u) {
        if (processedUnions.has(u))
            return;
        processedUnions = processedUnions.add(u);
        u.members.forEach(function (t) {
            if (t instanceof Type_1.UnionType) {
                addMembers(t);
            }
            else {
                members = members.add(t);
            }
        });
    }
    for (var _a = 0, unions_1 = unions; _a < unions_1.length; _a++) {
        var union = unions_1[_a];
        addMembers(union);
    }
    return members;
}
function flattenUnions(graph, stringTypeMapping, conflateNumbers) {
    var needsRepeat = false;
    function replace(types, builder, forwardingRef) {
        var unionBuilder = new UnifyClasses_1.UnifyUnionBuilder(builder, true, true, function (trefs, attributes) {
            Support_1.assert(trefs.length > 0, "Must have at least one type to build union");
            trefs = trefs.map(function (tref) { return builder.reconstituteType(tref.deref()[0]); });
            if (trefs.length === 1) {
                return trefs[0];
            }
            needsRepeat = true;
            return builder.getUnionType(attributes, immutable_1.OrderedSet(trefs));
        });
        return UnifyClasses_1.unifyTypes(types, immutable_1.Map(), builder, unionBuilder, conflateNumbers, forwardingRef);
    }
    var allUnions = graph.allTypesUnordered().filter(function (t) { return t instanceof Type_1.UnionType; });
    var nonCanonicalUnions = allUnions.filter(function (u) { return !u.isCanonical; });
    var singleTypeGroups = immutable_1.Map();
    var groups = [];
    var foundIntersection = false;
    nonCanonicalUnions.forEach(function (u) {
        var members = unionMembersRecursively(u);
        Support_1.assert(!members.isEmpty(), "We can't have an empty union");
        if (members.some(function (m) { return m instanceof Type_1.IntersectionType; })) {
            foundIntersection = true;
            return;
        }
        if (members.size === 1) {
            var t = Support_1.defined(members.first());
            var maybeSet = singleTypeGroups.get(t);
            if (maybeSet === undefined) {
                maybeSet = immutable_1.OrderedSet([t]);
            }
            maybeSet = maybeSet.add(u);
            singleTypeGroups = singleTypeGroups.set(t, maybeSet);
        }
        else {
            groups.push([u]);
        }
    });
    singleTypeGroups.forEach(function (ts) { return groups.push(ts.toArray()); });
    graph = graph.rewrite(stringTypeMapping, false, groups, replace);
    // console.log(`flattened ${nonCanonicalUnions.size} of ${unions.size} unions`);
    return [graph, !needsRepeat && !foundIntersection];
}
exports.flattenUnions = flattenUnions;
