"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Support_1 = require("./Support");
var TypeNames_1 = require("./TypeNames");
function isPrimitiveStringTypeKind(kind) {
    return kind === "string" || kind === "date" || kind === "time" || kind === "date-time";
}
exports.isPrimitiveStringTypeKind = isPrimitiveStringTypeKind;
function isNumberTypeKind(kind) {
    return kind === "integer" || kind === "double";
}
exports.isNumberTypeKind = isNumberTypeKind;
function isPrimitiveTypeKind(kind) {
    if (isPrimitiveStringTypeKind(kind))
        return true;
    if (isNumberTypeKind(kind))
        return true;
    return kind === "none" || kind === "any" || kind === "null" || kind === "bool";
}
exports.isPrimitiveTypeKind = isPrimitiveTypeKind;
function triviallyStructurallyCompatible(x, y) {
    if (x.typeRef.getIndex() === y.typeRef.getIndex())
        return true;
    if (x.kind === "none" || y.kind === "none")
        return true;
    return false;
}
var Type = /** @class */ (function () {
    function Type(typeRef, kind) {
        this.typeRef = typeRef;
        this.kind = kind;
    }
    Type.prototype.directlyReachableTypes = function (setForType) {
        var set = setForType(this);
        if (set)
            return set;
        return orderedSetUnion(this.children.map(function (t) { return t.directlyReachableTypes(setForType); }));
    };
    Type.prototype.getAttributes = function () {
        return this.typeRef.deref()[1];
    };
    Object.defineProperty(Type.prototype, "hasNames", {
        get: function () {
            return TypeNames_1.namesTypeAttributeKind.tryGetInAttributes(this.getAttributes()) !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    Type.prototype.getNames = function () {
        return Support_1.defined(TypeNames_1.namesTypeAttributeKind.tryGetInAttributes(this.getAttributes()));
    };
    Type.prototype.getCombinedName = function () {
        return this.getNames().combinedName;
    };
    Type.prototype.getProposedNames = function () {
        return this.getNames().proposedNames;
    };
    Type.prototype.equals = function (other) {
        if (!Object.prototype.hasOwnProperty.call(other, "typeRef")) {
            return false;
        }
        return this.typeRef.equals(other.typeRef);
    };
    Type.prototype.hashCode = function () {
        return this.typeRef.hashCode();
    };
    Type.prototype.structurallyCompatible = function (other) {
        if (triviallyStructurallyCompatible(this, other))
            return true;
        if (this.kind !== other.kind)
            return false;
        var workList = [[this, other]];
        // This contains a set of pairs which are the type pairs
        // we have already determined to be equal.  We can't just
        // do comparison recursively because types can have cycles.
        var done = [];
        var failed;
        var queue = function (x, y) {
            if (triviallyStructurallyCompatible(x, y))
                return true;
            if (x.kind !== y.kind) {
                failed = true;
                return false;
            }
            workList.push([x, y]);
            return true;
        };
        while (workList.length > 0) {
            var _a = Support_1.defined(workList.pop()), a = _a[0], b = _a[1];
            if (a.typeRef.getIndex() > b.typeRef.getIndex()) {
                _b = [b, a], a = _b[0], b = _b[1];
            }
            if (!a.isPrimitive()) {
                var ai = a.typeRef.getIndex();
                var bi = b.typeRef.getIndex();
                var found = false;
                for (var _i = 0, done_1 = done; _i < done_1.length; _i++) {
                    var _c = done_1[_i], dai = _c[0], dbi = _c[1];
                    if (dai === ai && dbi === bi) {
                        found = true;
                        break;
                    }
                }
                if (found)
                    continue;
                done.push([ai, bi]);
            }
            failed = false;
            if (!a.structuralEqualityStep(b, queue))
                return false;
            if (failed)
                return false;
        }
        return true;
        var _b;
    };
    Type.prototype.getParentTypes = function () {
        return this.typeRef.graph.getParentsOfType(this);
    };
    Type.prototype.getAncestorsNotInSet = function (set) {
        var workList = [this];
        var processed = immutable_1.Set();
        var ancestors = immutable_1.Set();
        for (;;) {
            var t = workList.pop();
            if (t === undefined)
                break;
            var parents = t.getParentTypes();
            console.log(parents.size + " parents");
            parents.forEach(function (p) {
                if (processed.has(p))
                    return;
                processed = processed.add(p);
                if (set.has(p.typeRef)) {
                    console.log("adding " + p.kind);
                    workList.push(p);
                }
                else {
                    console.log("found " + p.kind);
                    ancestors = ancestors.add(p);
                }
            });
        }
        return ancestors;
    };
    return Type;
}());
exports.Type = Type;
var PrimitiveType = /** @class */ (function (_super) {
    __extends(PrimitiveType, _super);
    function PrimitiveType(typeRef, kind, checkKind) {
        if (checkKind === void 0) { checkKind = true; }
        var _this = this;
        if (checkKind) {
            Support_1.assert(kind !== "string", "Cannot instantiate a PrimitiveType as string");
        }
        _this = _super.call(this, typeRef, kind) || this;
        return _this;
    }
    Object.defineProperty(PrimitiveType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(PrimitiveType.prototype, "isNullable", {
        get: function () {
            return this.kind === "null" || this.kind === "any" || this.kind === "none";
        },
        enumerable: true,
        configurable: true
    });
    PrimitiveType.prototype.isPrimitive = function () {
        return true;
    };
    PrimitiveType.prototype.map = function (builder, _) {
        // console.log(`${mapIndentation()}mapping ${this.kind}`);
        return builder.getPrimitiveType(this.kind);
    };
    PrimitiveType.prototype.structuralEqualityStep = function (_other, _queue) {
        return true;
    };
    return PrimitiveType;
}(Type));
exports.PrimitiveType = PrimitiveType;
function isNull(t) {
    return t.kind === "null";
}
var StringType = /** @class */ (function (_super) {
    __extends(StringType, _super);
    function StringType(typeRef, enumCases) {
        var _this = _super.call(this, typeRef, "string", false) || this;
        _this.enumCases = enumCases;
        return _this;
    }
    StringType.prototype.map = function (builder, _) {
        // console.log(`${mapIndentation()}mapping ${this.kind}`);
        return builder.getStringType(this.enumCases);
    };
    StringType.prototype.structuralEqualityStep = function (_other, _queue) {
        return true;
    };
    return StringType;
}(PrimitiveType));
exports.StringType = StringType;
var ArrayType = /** @class */ (function (_super) {
    __extends(ArrayType, _super);
    function ArrayType(typeRef, _itemsRef) {
        var _this = _super.call(this, typeRef, "array") || this;
        _this._itemsRef = _itemsRef;
        return _this;
    }
    ArrayType.prototype.setItems = function (itemsRef) {
        if (this._itemsRef !== undefined) {
            return Support_1.panic("Can only set array items once");
        }
        this._itemsRef = itemsRef;
    };
    ArrayType.prototype.getItemsRef = function () {
        if (this._itemsRef === undefined) {
            return Support_1.panic("Array items accessed before they were set");
        }
        return this._itemsRef;
    };
    Object.defineProperty(ArrayType.prototype, "items", {
        get: function () {
            return this.getItemsRef().deref()[0];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ArrayType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.items]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ArrayType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ArrayType.prototype.isPrimitive = function () {
        return false;
    };
    ArrayType.prototype.map = function (builder, f) {
        // console.log(`${mapIndentation()}mapping ${this.kind}`);
        // mapPath.push("[]");
        var result = builder.getArrayType(f(this.getItemsRef()));
        // mapPath.pop();
        return result;
    };
    ArrayType.prototype.structuralEqualityStep = function (other, queue) {
        return queue(this.items, other.items);
    };
    return ArrayType;
}(Type));
exports.ArrayType = ArrayType;
var MapType = /** @class */ (function (_super) {
    __extends(MapType, _super);
    function MapType(typeRef, _valuesRef) {
        var _this = _super.call(this, typeRef, "map") || this;
        _this._valuesRef = _valuesRef;
        return _this;
    }
    MapType.prototype.setValues = function (valuesRef) {
        if (this._valuesRef !== undefined) {
            return Support_1.panic("Can only set map values once");
        }
        this._valuesRef = valuesRef;
    };
    MapType.prototype.getValuesRef = function () {
        if (this._valuesRef === undefined) {
            return Support_1.panic("Map values accessed before they were set");
        }
        return this._valuesRef;
    };
    Object.defineProperty(MapType.prototype, "values", {
        get: function () {
            return this.getValuesRef().deref()[0];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MapType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.values]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MapType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    MapType.prototype.isPrimitive = function () {
        return false;
    };
    MapType.prototype.map = function (builder, f) {
        // console.log(`${mapIndentation()}mapping ${this.kind}`);
        // mapPath.push("{}");
        var result = builder.getMapType(f(this.getValuesRef()));
        // mapPath.pop();
        return result;
    };
    MapType.prototype.structuralEqualityStep = function (other, queue) {
        return queue(this.values, other.values);
    };
    return MapType;
}(Type));
exports.MapType = MapType;
var GenericClassProperty = /** @class */ (function () {
    function GenericClassProperty(typeData, isOptional) {
        this.typeData = typeData;
        this.isOptional = isOptional;
    }
    GenericClassProperty.prototype.equals = function (other) {
        if (!(other instanceof GenericClassProperty)) {
            return false;
        }
        return immutable_1.is(this.typeData, other.typeData) && this.isOptional === other.isOptional;
    };
    GenericClassProperty.prototype.hashCode = function () {
        return immutable_1.hash(this.typeData) + (this.isOptional ? 17 : 23);
    };
    return GenericClassProperty;
}());
exports.GenericClassProperty = GenericClassProperty;
var ClassProperty = /** @class */ (function (_super) {
    __extends(ClassProperty, _super);
    function ClassProperty(typeRef, isOptional) {
        return _super.call(this, typeRef, isOptional) || this;
    }
    Object.defineProperty(ClassProperty.prototype, "typeRef", {
        get: function () {
            return this.typeData;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassProperty.prototype, "type", {
        get: function () {
            return this.typeRef.deref()[0];
        },
        enumerable: true,
        configurable: true
    });
    return ClassProperty;
}(GenericClassProperty));
exports.ClassProperty = ClassProperty;
function propertiesAreSorted(_props) {
    /*
    const keys = props.keySeq().toArray();
    for (let i = 1; i < keys.length; i++) {
        if (keys[i - 1] > keys[i]) return false;
    }
*/
    return true;
}
var ClassType = /** @class */ (function (_super) {
    __extends(ClassType, _super);
    function ClassType(typeRef, isFixed, _properties) {
        var _this = _super.call(this, typeRef, "class") || this;
        _this.isFixed = isFixed;
        _this._properties = _properties;
        if (_properties !== undefined) {
            Support_1.assert(propertiesAreSorted(_properties));
        }
        return _this;
    }
    ClassType.prototype.setProperties = function (properties) {
        Support_1.assert(propertiesAreSorted(properties));
        if (this._properties !== undefined) {
            return Support_1.panic("Can only set class properties once");
        }
        this._properties = properties;
    };
    Object.defineProperty(ClassType.prototype, "properties", {
        get: function () {
            if (this._properties === undefined) {
                return Support_1.panic("Properties are not set yet");
            }
            return this._properties;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "sortedProperties", {
        get: function () {
            var properties = this.properties;
            var sortedKeys = properties.keySeq().sort();
            var props = sortedKeys.map(function (k) { return [k, Support_1.defined(properties.get(k))]; });
            return immutable_1.OrderedMap(props);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "children", {
        get: function () {
            return this.sortedProperties.map(function (p) { return p.type; }).toOrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ClassType.prototype.isPrimitive = function () {
        return false;
    };
    ClassType.prototype.map = function (builder, f) {
        // const indent = "  ".repeat(mapPath.length);
        // const path = mapPath.join(".");
        // console.log(`${indent} mapping class ${this.getCombinedName()} at ${path}`);
        var properties = this.properties.map(function (p, _n) {
            // console.log(`${indent}  property ${path}.${n}`);
            // mapPath.push(n);
            var result = new ClassProperty(f(p.typeRef), p.isOptional);
            // mapPath.pop();
            return result;
        });
        if (this.isFixed) {
            return builder.getUniqueClassType(true, properties);
        }
        else {
            return builder.getClassType(properties);
        }
    };
    ClassType.prototype.structuralEqualityStep = function (other, queue) {
        var pa = this.properties;
        var pb = other.properties;
        if (pa.size !== pb.size)
            return false;
        var failed = false;
        pa.forEach(function (cpa, name) {
            var cpb = pb.get(name);
            if (cpb === undefined || cpa.isOptional !== cpb.isOptional || !queue(cpa.type, cpb.type)) {
                failed = true;
                return false;
            }
        });
        return !failed;
    };
    return ClassType;
}(Type));
exports.ClassType = ClassType;
function assertIsClass(t) {
    if (!(t instanceof ClassType)) {
        return Support_1.panic("Supposed class type is not a class type");
    }
    return t;
}
exports.assertIsClass = assertIsClass;
var EnumType = /** @class */ (function (_super) {
    __extends(EnumType, _super);
    function EnumType(typeRef, cases) {
        var _this = _super.call(this, typeRef, "enum") || this;
        _this.cases = cases;
        return _this;
    }
    Object.defineProperty(EnumType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(EnumType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    EnumType.prototype.isPrimitive = function () {
        return false;
    };
    EnumType.prototype.map = function (builder, _) {
        // console.log(`${mapIndentation()}mapping ${this.kind}`);
        return builder.getEnumType(this.cases);
    };
    EnumType.prototype.structuralEqualityStep = function (other, _queue) {
        return this.cases.toSet().equals(other.cases.toSet());
    };
    return EnumType;
}(Type));
exports.EnumType = EnumType;
var SetOperationType = /** @class */ (function (_super) {
    __extends(SetOperationType, _super);
    function SetOperationType(typeRef, kind, _memberRefs) {
        var _this = _super.call(this, typeRef, kind) || this;
        _this._memberRefs = _memberRefs;
        return _this;
    }
    SetOperationType.prototype.setMembers = function (memberRefs) {
        if (this._memberRefs !== undefined) {
            return Support_1.panic("Can only set map members once");
        }
        this._memberRefs = memberRefs;
    };
    SetOperationType.prototype.getMemberRefs = function () {
        if (this._memberRefs === undefined) {
            return Support_1.panic("Map members accessed before they were set");
        }
        return this._memberRefs;
    };
    Object.defineProperty(SetOperationType.prototype, "members", {
        get: function () {
            return this.getMemberRefs().map(function (tref) { return tref.deref()[0]; });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SetOperationType.prototype, "sortedMembers", {
        get: function () {
            // FIXME: We're assuming no two members of the same kind.
            return this.members.sortBy(function (t) { return t.kind; });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SetOperationType.prototype, "children", {
        get: function () {
            return this.sortedMembers;
        },
        enumerable: true,
        configurable: true
    });
    SetOperationType.prototype.isPrimitive = function () {
        return false;
    };
    SetOperationType.prototype.structuralEqualityStep = function (other, queue) {
        return setOperationCasesEqual(this.members, other.members, queue);
    };
    return SetOperationType;
}(Type));
exports.SetOperationType = SetOperationType;
var IntersectionType = /** @class */ (function (_super) {
    __extends(IntersectionType, _super);
    function IntersectionType(typeRef, memberRefs) {
        return _super.call(this, typeRef, "intersection", memberRefs) || this;
    }
    Object.defineProperty(IntersectionType.prototype, "isNullable", {
        get: function () {
            return Support_1.panic("isNullable not implemented for IntersectionType");
        },
        enumerable: true,
        configurable: true
    });
    IntersectionType.prototype.map = function (builder, f) {
        var members = this.getMemberRefs().map(f);
        // FIXME: Eventually switch to non-unique
        return builder.getUniqueIntersectionType(members);
    };
    return IntersectionType;
}(SetOperationType));
exports.IntersectionType = IntersectionType;
var UnionType = /** @class */ (function (_super) {
    __extends(UnionType, _super);
    function UnionType(typeRef, memberRefs) {
        return _super.call(this, typeRef, "union", memberRefs) || this;
    }
    Object.defineProperty(UnionType.prototype, "stringTypeMembers", {
        get: function () {
            return this.members.filter(function (t) { return ["string", "date", "time", "date-time", "enum"].indexOf(t.kind) >= 0; });
        },
        enumerable: true,
        configurable: true
    });
    UnionType.prototype.findMember = function (kind) {
        return this.members.find(function (t) { return t.kind === kind; });
    };
    Object.defineProperty(UnionType.prototype, "isNullable", {
        get: function () {
            return this.findMember("null") !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "isCanonical", {
        get: function () {
            var members = this.members;
            if (members.size <= 1)
                return false;
            var kinds = members.map(function (t) { return t.kind; });
            if (kinds.size < members.size)
                return false;
            if (kinds.has("union") || kinds.has("intersection"))
                return false;
            if (kinds.has("none") || kinds.has("any"))
                return false;
            if (kinds.has("string") && kinds.has("enum"))
                return false;
            if (kinds.has("class") && kinds.has("map"))
                return false;
            return true;
        },
        enumerable: true,
        configurable: true
    });
    UnionType.prototype.map = function (builder, f) {
        var members = this.getMemberRefs().map(f);
        return builder.getUnionType(members);
    };
    return UnionType;
}(SetOperationType));
exports.UnionType = UnionType;
function setOperationCasesEqual(ma, mb, membersEqual) {
    if (ma.size !== mb.size)
        return false;
    var failed = false;
    ma.forEach(function (ta) {
        var tb = mb.find(function (t) { return t.kind === ta.kind; });
        if (tb === undefined || !membersEqual(ta, tb)) {
            failed = true;
            return false;
        }
    });
    return !failed;
}
exports.setOperationCasesEqual = setOperationCasesEqual;
// FIXME: We shouldn't have to sort here.  This is just because we're not getting
// back the right order from JSON Schema, due to the changes the intersection types
// introduced.
function removeNullFromUnion(t, sortBy) {
    if (sortBy === void 0) { sortBy = false; }
    function sort(s) {
        if (sortBy === false)
            return s;
        if (sortBy === true)
            return s.sortBy(function (m) { return m.kind; });
        return s.sortBy(sortBy);
    }
    var nullType = t.findMember("null");
    if (nullType === undefined) {
        return [null, sort(t.members)];
    }
    return [nullType, sort(t.members.filterNot(isNull))];
}
exports.removeNullFromUnion = removeNullFromUnion;
function removeNullFromType(t) {
    if (t.kind === "null") {
        return [t, immutable_1.OrderedSet()];
    }
    if (!(t instanceof UnionType)) {
        return [null, immutable_1.OrderedSet([t])];
    }
    return removeNullFromUnion(t);
}
exports.removeNullFromType = removeNullFromType;
function nullableFromUnion(t) {
    var _a = removeNullFromUnion(t), hasNull = _a[0], nonNulls = _a[1];
    if (hasNull === null)
        return null;
    if (nonNulls.size !== 1)
        return null;
    return Support_1.defined(nonNulls.first());
}
exports.nullableFromUnion = nullableFromUnion;
function nonNullTypeCases(t) {
    return removeNullFromType(t)[1];
}
exports.nonNullTypeCases = nonNullTypeCases;
function getNullAsOptional(cp) {
    var _a = removeNullFromType(cp.type), maybeNull = _a[0], nonNulls = _a[1];
    if (cp.isOptional) {
        return [true, nonNulls];
    }
    return [maybeNull !== null, nonNulls];
}
exports.getNullAsOptional = getNullAsOptional;
// FIXME: The outer OrderedSet should be some Collection, but I can't figure out
// which one.  Collection.Indexed doesn't work with OrderedSet, which is unfortunate.
function orderedSetUnion(sets) {
    var setArray = sets.toArray();
    if (setArray.length === 0)
        return immutable_1.OrderedSet();
    if (setArray.length === 1)
        return setArray[0];
    return (_a = setArray[0]).union.apply(_a, setArray.slice(1));
    var _a;
}
// FIXME: Give this an appropriate name, considering that we don't distinguish
// between named and non-named types anymore.
function isNamedType(t) {
    return t instanceof ClassType || t instanceof EnumType || t instanceof UnionType;
}
exports.isNamedType = isNamedType;
function separateNamedTypes(types) {
    var classes = types.filter(function (t) { return t instanceof ClassType; }).toOrderedSet();
    var enums = types.filter(function (t) { return t instanceof EnumType; }).toOrderedSet();
    var unions = types.filter(function (t) { return t instanceof UnionType; }).toOrderedSet();
    return { classes: classes, enums: enums, unions: unions };
}
exports.separateNamedTypes = separateNamedTypes;
function directlyReachableSingleNamedType(type) {
    var definedTypes = type.directlyReachableTypes(function (t) {
        if ((!(t instanceof UnionType) && isNamedType(t)) ||
            (t instanceof UnionType && nullableFromUnion(t) === null)) {
            return immutable_1.OrderedSet([t]);
        }
        return null;
    });
    Support_1.assert(definedTypes.size <= 1, "Cannot have more than one defined type per top-level");
    return definedTypes.first();
}
exports.directlyReachableSingleNamedType = directlyReachableSingleNamedType;
function matchTypeExhaustive(t, noneType, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, dateType, timeType, dateTimeType) {
    if (t.isPrimitive()) {
        var f = {
            none: noneType,
            any: anyType,
            null: nullType,
            bool: boolType,
            integer: integerType,
            double: doubleType,
            string: stringType,
            date: dateType,
            time: timeType,
            "date-time": dateTimeType
        }[t.kind];
        if (f !== undefined)
            return f(t);
        return Support_1.assertNever(f);
    }
    else if (t instanceof ArrayType)
        return arrayType(t);
    else if (t instanceof ClassType)
        return classType(t);
    else if (t instanceof MapType)
        return mapType(t);
    else if (t instanceof EnumType)
        return enumType(t);
    else if (t instanceof UnionType)
        return unionType(t);
    return Support_1.panic("Unknown Type");
}
exports.matchTypeExhaustive = matchTypeExhaustive;
function matchType(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, stringTypeMatchers) {
    function typeNotSupported(_) {
        return Support_1.panic("Unsupported PrimitiveType");
    }
    if (stringTypeMatchers === undefined) {
        stringTypeMatchers = {};
    }
    /* tslint:disable:strict-boolean-expressions */
    return matchTypeExhaustive(t, typeNotSupported, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, stringTypeMatchers.dateType || typeNotSupported, stringTypeMatchers.timeType || typeNotSupported, stringTypeMatchers.dateTimeType || typeNotSupported);
    /* tslint:enable */
}
exports.matchType = matchType;
function matchCompoundType(t, arrayType, classType, mapType, unionType) {
    function ignore(_) {
        return;
    }
    return matchType(t, ignore, ignore, ignore, ignore, ignore, ignore, arrayType, classType, mapType, ignore, unionType, { dateType: ignore, timeType: ignore, dateTimeType: ignore });
}
exports.matchCompoundType = matchCompoundType;
